from io import BytesIO
import logging
from typing import List, Optional

import numpy as np

from .aggregator import AggregatorABC

LOG = logging.getLogger("PythonModelSumAggregator")


class Aggregator(AggregatorABC):
    def __init__(self) -> None:
        LOG.info("initializing aggregator")
        self.global_weights: np.ndarray = None
        self.weights: List[np.ndarray] = []

    def add_weights(self, data: bytes) -> bool:
        LOG.info("adding weights (len = %d)", len(data))
        reader = BytesIO(data)
        weights = np.load(reader, allow_pickle=False)
        self.weights.append(weights)
        return True

    def aggregate(self) -> bytes:
        LOG.info("starting aggregation (%d models)", len(self.weights))
        self.global_weights = sum(self.weights)
        LOG.info("finished aggregation")
        return self.get_global_weights()

    def reset(self, global_weights: Optional[bytes]) -> None:
        LOG.info("resetting aggregator")
        if global_weights is None:
            self.global_weights = None
        else:
            reader = BytesIO(global_weights)
            self.global_weights = np.load(reader)
        self.weights = []

    def get_global_weights(self) -> bytes:
        LOG.info("returning global weights")
        writer = BytesIO()
        np.save(writer, self.global_weights, allow_pickle=False)
        # We cannot use getvalue here because it copies the buffer.
        # getbuffer will not as long as the data is not modified.
        return writer.getbuffer()[:]



from abc import ABC, abstractmethod
from typing import Optional


class AggregatorABC(ABC):
    @abstractmethod
    def aggregate(self) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def add_weights(self, data: bytes) -> bool:
        raise NotImplementedError()

    @abstractmethod
    def reset(self, global_weights: Optional[bytes]) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get_global_weights(self) -> bytes:
        raise NotImplementedError()



from io import BytesIO
import logging
from typing import List, Optional

import numpy as np

from .aggregator import AggregatorABC

LOG = logging.getLogger("PythonWeightedAverageAggregator")


class Aggregator(AggregatorABC):
    def __init__(self) -> None:
        logging.basicConfig(level=logging.INFO)
        LOG.info("initializing aggregator")
        self.global_weights: np.ndarray = None
        self.weights: List[np.ndarray] = []
        self.aggregation_data: List[int] = []

    def add_weights(self, data: bytes) -> bool:
        LOG.info("adding weights (len = %d)", len(data))

        number_of_samples = int.from_bytes(data[:4], byteorder="big")
        weights = np.load(BytesIO(data[4:]), allow_pickle=False)

        self.aggregation_data.append(number_of_samples)
        self.weights.append(weights)

        return True

    def aggregate(self) -> bytes:
        LOG.info("starting aggregation (%d models)", len(self.weights))
        aggregation_weights: np.ndarray
        if any(self.aggregation_data):
            aggregation_weights = np.array(self.aggregation_data) / np.sum(
                self.aggregation_data
            )
        else:
            aggregation_weights = np.ones_like(self.aggregation_data) / len(
                self.aggregation_data
            )

        scaled_model_weights: List[np.ndarray] = [
            model_weight * aggregation_weight
            for model_weight, aggregation_weight in zip(
                self.weights, aggregation_weights
            )
        ]
        self.global_weights = np.sum(scaled_model_weights, axis=0)
        # If global_weights is a scalar, make it a one dimensional
        # array
        if self.global_weights.shape == ():
            self.global_weights = np.array([self.global_weights])
        self.weights = []
        self.aggregation_data = []
        LOG.info("finished aggregation")
        return self.get_global_weights()

    def reset(self, global_weights: Optional[bytes]) -> None:
        LOG.info("resetting aggregator")
        if global_weights is None:
            self.global_weights = None
        else:
            reader = BytesIO(global_weights)
            self.global_weights = np.load(reader, allow_pickle=False)

        self.weights = []
        self.aggregation_data = []

    def get_global_weights(self) -> bytes:
        LOG.info("returning global weights")
        if self.global_weights is None:
            return b""
        writer = BytesIO()
        np.save(writer, self.global_weights, allow_pickle=False)
        # We cannot use getvalue here because it copies the buffer.
        # getbuffer will not as long as the data is not modified.
        return writer.getbuffer()[:]



# pylint: disable=exec-used,invalid-name
import os.path

from setuptools import find_packages, setup

project_dir = os.path.dirname(os.path.abspath(__file__))
version_file_path = os.path.join(project_dir, "xain_aggregators/__version__.py")
readme_file_path = os.path.join(project_dir, "README.md")

version = {}
with open(version_file_path) as fp:
    exec(fp.read(), version)

with open(readme_file_path, "r") as fp:
    readme = fp.read()

install_requires = ["numpy"]

dev_require = [
    "black",
    "mypy",
    "pylint",
    "isort",
    "pip-licenses",
]

tests_require = [
    "pytest",
]

docs_require = []

setup(
    name="xain_aggregators",
    version=version["__version__"],
    description="XAIN is an open source framework for federated learning.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/xainag/xain-sdk",
    author=["XAIN AG"],
    author_email="services@xain.io",
    license="Apache License Version 2.0",
    python_requires=">=3.6",
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        "test": tests_require,
        "docs": docs_require,
        "dev": dev_require + tests_require + docs_require,
    },
)
